from collections import OrderedDict
from operator import add
from typing import List, Optional

from .config import Config
from .loader import ConfigLoader
from .snippet import ConfigSnippet
from .types import PROFILE_GLOBAL
from .utils import list_flatten, dict_merge_with_wildcard


class ConfigSet(OrderedDict):
    @classmethod
    def load(cls, *args, **kwargs) -> "ConfigSet":
        loader = ConfigLoader(*args, **kwargs)
        loader.load()

        snippets = list_flatten(loader.values())

        def name(snippet: ConfigSnippet, from_fields: List[str]):
            ids = [
                n
                for n in [snippet.find(field) for field in from_fields]
                if n is not None
            ]
            if ids:
                return "-".join(ids)
            else:
                return PROFILE_GLOBAL

        named_snippets = dict()
        for snippet in snippets:
            named_snippets.setdefault(name(snippet, ["group", "name"]), []).append(
                snippet
            )

        named_configs = {
            name: Config.from_snippets(snippets)
            for name, snippets in named_snippets.items()
        }
        return cls(named_configs)

    def get(self, key: str, default: Optional[Config] = None) -> Optional[Config]:
        return OrderedDict.get(self, key, default)

    def __add__(self, other):
        return dict_merge_with_wildcard(self, other, add)



from abc import ABC, abstractmethod
from collections import OrderedDict
from os import path

import yaml

from .snippet import ConfigSnippet
from .source import ConfigSource


class BaseConfigLoader(OrderedDict, ABC):
    """
    Key: the path of loaded yaml
    Value: the config dict of yaml
    """

    def __init__(self, *files: str):
        self._source = files
        self._includes = []

    @abstractmethod
    def _parse_config(self, content):
        ...

    def find(self, file, key):
        for snippet in self.get(file):
            v = snippet.get(key, None)
            if v is not None:
                return v
        return None

    def load(self):
        [self._load(file) for file in self._source]

    def _read_file(self, file):
        key = path.abspath(file)
        if key in self.keys():
            return None
        with open(file, "r") as fh:
            return fh.read()

    def _load(self, file: str, source=None) -> None:
        file = path.abspath(file)
        _conf = self._parse_config(self._read_file(file))

        configs = [self._process_directives(file, c) for c in _conf]
        configs = [dc for dc in configs if dc]

        def gs(config, file, index, ps=None):
            source = ConfigSource(uri=file, index=index, loader=self)
            if ps is not None:
                source = ps + source
            return ConfigSnippet(config=config, source=source)

        snippets = [
            gs(config, file, index, source) for index, config in enumerate(configs)
        ]
        self.setdefault(file, snippets)

        while self._includes:
            inc, source = self._includes.pop()
            self._load(inc, source)

    def _process_directives(self, file, config: dict) -> dict:
        workdir = path.dirname(file)
        includes = config.pop("@include", None)
        if includes is not None:
            self._includes.extend(
                [
                    (
                        path.abspath(path.join(workdir, f)),
                        ConfigSource(uri=file, index=0, loader=self),
                    )
                    for f in includes
                ]
            )

        return config


class YamlConfigLoader(BaseConfigLoader):
    def _parse_config(self, content):
        if content is None:
            return []
        return yaml.load_all(content, Loader=yaml.SafeLoader)


class ConfigLoader(object):
    def __new__(cls, *args, **kwargs):
        return YamlConfigLoader(*args, **kwargs)



from collections import OrderedDict
from functools import reduce
from operator import add
from typing import List, Optional

from .snippet import ConfigSnippet
from .types import PROFILE_GLOBAL, ProfileKey
from .utils import list_groupby, dict_merge_with_wildcard


class Config(OrderedDict):
    name: str

    @classmethod
    def from_snippets(cls, snippets: List[ConfigSnippet]) -> "Config":
        def groupby_profile_and_merge(
            snippets: List[ConfigSnippet],
        ) -> List[ConfigSnippet]:
            return [reduce(add, g) for g in list_groupby(snippets, lambda s: s.profile)]

        gs = groupby_profile_and_merge(snippets)
        config = OrderedDict()
        for snippet in gs:
            config.setdefault(snippet.profile, snippet)

        global_profile = config.pop(PROFILE_GLOBAL, None)
        if global_profile is not None:
            config = {p: global_profile + c for p, c in config.items()}
            config.update({PROFILE_GLOBAL: global_profile})

        return Config(config)

    def get(
        self, key: str, default: Optional[ConfigSnippet] = None
    ) -> Optional[ConfigSnippet]:
        return super().get(key, default)

    def profile(self, key: ProfileKey):
        return self.get(key, self.get(PROFILE_GLOBAL))

    def __add__(self, other: "Config") -> "Config":
        return dict_merge_with_wildcard(self, other, add)



from collections import OrderedDict
from copy import deepcopy, copy
from functools import singledispatch
from itertools import groupby
from typing import Callable, Dict, List, Any
import pprint
from .types import KT, VT, PROFILE_GLOBAL


def inspect(obj):
    pp = pprint.PrettyPrinter()
    pp.pprint(obj)


def list_flatten(t: list):
    return [item for sublist in t for item in sublist]


def list_groupby(iterable: List[Any], projection) -> List[List[Any]]:
    return [list(it) for k, it in groupby(sorted(iterable, key=projection), projection)]


def list_uniq(a: List) -> List:
    return list(dict.fromkeys(a))


def dict_merge(
    a: Dict[KT, VT], b: Dict[KT, VT], f: Callable[[VT, VT], VT]
) -> Dict[KT, VT]:
    merged = copy(a)
    for k in b.keys():
        if k in merged.keys():
            merged[k] = f(a[k], b[k])
        else:
            merged[k] = b[k]
    return merged


def dict_merge_with_wildcard(a, b, f):
    if b is None:
        return a

    merged = copy(b)
    wb = b.get(PROFILE_GLOBAL, None)
    for k in a.keys():
        if k in b.keys():
            merged[k] = f(a[k], b[k])
        else:
            merged[k] = f(a[k], wb)

    wa = a.get(PROFILE_GLOBAL, None)
    if wa is not None:
        for k in b.keys():
            if k not in a.keys():
                merged[k] = f(wa, b[k])
    return merged


@singledispatch
def config_merger(_, b):
    return b


@config_merger.register
def _(a: dict, b: dict):
    return dict_merge(a, b, config_merger)


@config_merger.register
def _(a: list, b: list):
    return a + b



from collections import OrderedDict
from copy import copy, deepcopy
from typing import Dict, Any, Callable

from .source import ConfigSource
from .types import PROFILE_GLOBAL
from .utils import dict_merge, config_merger


class ConfigSnippet(OrderedDict):
    def __init__(self, config: Dict[str, Any], source: ConfigSource):
        super().__init__(config)
        self.source = source

    @property
    def profile(
        self,
        getter: Callable[["ConfigSnippet"], str] = lambda x: x.get(
            "profile", PROFILE_GLOBAL
        ),
    ):
        return getter(self)

    def __add__(self, other: "ConfigSnippet"):
        if other is None:
            return copy(self)

        return ConfigSnippet(
            config=dict_merge(self, other, config_merger),
            source=self.source + other.source,
        )

    def __copy__(self):
        return ConfigSnippet(
            config={k: copy(v) for k, v in self.items()}, source=self.source
        )

    def __deepcopy__(self, memodict={}):
        return ConfigSnippet(
            config={k: deepcopy(v, memodict) for k, v in self.items()},
            source=self.source,
        )

    def find(self, key):
        return self.get(key, self.source.find(key))
